/*
  * PmMediaGstAdPcmDecLib.c: Palm Media gStreamer IMA ADPCM Decoder Element
  *
  * Copyright 2009 Palm, Inc. All rights reserved.
  *
  * This program is free software and licensed under the terms of the GNU
  * Lesser General Public License Version 2.1 as published by the Free
  * Software Foundation;
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU Lesser General Public License for more details.

  * You should have received a copy of the GNU Lesser General Public License,
  * Version 2.1 along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-
  * 1301, USA
  */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <gst/gst.h>
#include <glib.h>

#include <Arm/Libraries/PmMediaAdPcmCodecLib/PmMediaAdPcmCodecLib.h>
#include "PmMediaGstAdPcmDecLib.h"

GST_DEBUG_CATEGORY_STATIC (gst_palmadpcm_debug);
#define GST_CAT_DEFAULT gst_palmadpcm_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0,
  ARG_SILENT
};

#define SINK_CAPS \
  "audio/x-adpcm, "               \
  "layout = (string) dvi, " \
  "block_align = (int) [ 1, 8192 ], "       \
  "rate = (int) [8000, 48000 ], "            \
  "channels = (int) [ 1, 2],"              \
  "width = (int) [8, 16], "\
  "depth = (int) [8, 16]" \

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(SINK_CAPS)
   );

#define SRC_CAPS_1 \
  "audio/x-raw-int, "               \
  "endianness = (int) BYTE_ORDER, " \
  "signed = (boolean) true, "       \
  "width = (int) 16, "              \
  "depth = (int) 16, "              \
  "rate = (int) [8000, 48000 ], "            \
  "channels = (int) [ 1, 2]"              \

#define SRC_CAPS_2 \
  "audio/x-raw-int, "               \
  "rate = (int) 11025, "            \
  "channels = (int) 2 "              \

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(SRC_CAPS_1)
    );

GST_BOILERPLATE (GstpalmADPCM, gst_palmadpcm, GstElement,
    GST_TYPE_ELEMENT);

static void gst_palmadpcm_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_palmadpcm_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_palmadpcm_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_palmadpcm_chain (GstPad * pad, GstBuffer * buf);
static GstStateChangeReturn gst_palmadpcm_change_state (GstElement * element,
    						       GstStateChange transition);

// static void gst_palmadpcm_type_find (GstTypeFind* tf, gpointer data);

static void
gst_palmadpcm_base_init (gpointer gclass)
{
  static GstElementDetails element_details = GST_ELEMENT_DETAILS(
    "ADPCM Decoder",
    "Codec/Decoder/Audio",
    "Decode an IMA/ADPCM file",
    "Palm Media Division"
  );

  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
  gst_element_class_set_details (element_class, &element_details);
}

/* initialize the plugin's class */
static void
gst_palmadpcm_class_init (GstpalmADPCMClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gstelement_class->change_state = gst_palmadpcm_change_state;

  gobject_class->set_property = gst_palmadpcm_set_property;
  gobject_class->get_property = gst_palmadpcm_get_property;

  g_object_class_install_property (gobject_class, ARG_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_palmadpcm_init (GstpalmADPCM * filter,
    GstpalmADPCMClass * gclass)
{
  GstElementClass *klass = GST_ELEMENT_GET_CLASS (filter);

  filter->sinkpad =
      gst_pad_new_from_template (gst_element_class_get_pad_template (klass,
          "sink"), "sink");
  gst_pad_set_setcaps_function (filter->sinkpad, gst_palmadpcm_set_caps);
//  gst_pad_set_getcaps_function (filter->sinkpad, gst_pad_proxy_getcaps);

  filter->srcpad =
      gst_pad_new_from_template (gst_element_class_get_pad_template (klass,
          "src"), "src");
//  gst_pad_set_getcaps_function (filter->srcpad, gst_pad_proxy_getcaps);

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_pad_set_chain_function (filter->sinkpad, gst_palmadpcm_chain);

  filter->silent = FALSE;
  filter->codec_initialized = FALSE;
  filter->bRunning = FALSE;
}

static GstStateChangeReturn
gst_palmadpcm_change_state (GstElement * element, GstStateChange transition)
{
  GstStateChangeReturn ret;
  GstpalmADPCM *filter = GST_PALMADPCM (element);

  GST_LOG("change_state\n");
  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      GST_LOG("NULL to READY\n");
      break;
    case GST_STATE_CHANGE_READY_TO_PAUSED:
      GST_LOG("READY to PAUSED\n");
      filter->bRunning = TRUE;
      break;
    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
      GST_LOG("PAUSED to PLAYING\n");
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);

  switch (transition) {
    case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
      GST_LOG("2: PLAYING to PAUSED\n");
      filter->bRunning = FALSE;
      break;
    case GST_STATE_CHANGE_PAUSED_TO_READY:
      GST_LOG("2: PAUSED to READY\n");
      break;
    case GST_STATE_CHANGE_READY_TO_NULL:
      GST_LOG("2: READY to NULL\n");
      break;
    default:
      break;
  }
  return ret;
}

static void
gst_palmadpcm_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstpalmADPCM *filter = GST_PALMADPCM (object);

  switch (prop_id) {
    case ARG_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_palmadpcm_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstpalmADPCM *filter = GST_PALMADPCM (object);

  switch (prop_id) {
    case ARG_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_palmadpcm_set_caps (GstPad * pad, GstCaps * caps)
{
  GstpalmADPCM *filter;
  GstStructure *ins;
  GstCaps *base_caps;
  const GValue *val;
  GstBuffer *pData;
  gchar *data;
  gint size;

  filter = GST_PALMADPCM (gst_pad_get_parent (pad));

  ins = gst_caps_get_structure(caps, 0);

  if(gst_structure_has_name (ins, "audio/x-adpcm"))
	printf("It's got ADPCM\n");

  if (gst_structure_get_int (ins, "channels", &filter->channels)) {
	printf("channels: %d\n", filter->channels);
  }

  if (gst_structure_get_int (ins, "rate", &filter->rate)) {
	printf("rate: %d\n", filter->rate);
  }

  if (gst_structure_get_int (ins, "block_align", &filter->block_align)) {
	printf("block_align: %d\n", filter->block_align);
  }

  filter->adpcm_type = 0x0;		// Assume BAD

  if ((data = (gchar *)gst_structure_get_string(ins, "layout")) != 0) {
	printf("layout %s\n", data);
	if(strncmp(data, "dvi", 3) == 0)
		filter->adpcm_type = sndFormatDVI_ADPCM;
	if(strncmp(data, "microsoft", 9) == 0)
		filter->adpcm_type = sndFormatIMA_ADPCM;
  }

  val = gst_structure_get_value (ins, "codec_data");
  if (val == NULL) {
	printf("Crap, no codec_data\n");
  } else {
	pData = (GstBuffer *) gst_value_get_mini_object (val);
	g_return_val_if_fail (pData != NULL, FALSE);
	gst_buffer_ref (pData);
	data = (gchar *) GST_BUFFER_DATA (pData);
	size = GST_BUFFER_SIZE (pData);
	printf("%d bytes of data in %p\n", size, data);
	filter->pExtraData = (unsigned char *)malloc(size);
	memcpy(filter->pExtraData, data, size);
	printf("%x %x\n", data[0], data[1]);
  }

  base_caps = gst_caps_new_simple("audio/x-raw-int","rate", G_TYPE_INT, 
	filter->rate, "channels", G_TYPE_INT, filter->channels, 
	   "width", G_TYPE_INT, 16, 
	   "depth", G_TYPE_INT, 16, 
	   "endianness", G_TYPE_INT, G_BYTE_ORDER,
	   "signed", G_TYPE_BOOLEAN, TRUE,
	   NULL);

  gst_pad_set_caps (filter->srcpad, base_caps);
  gst_caps_unref (base_caps);

  gst_object_unref (GST_OBJECT(filter));
  
  return TRUE;

//  return gst_pad_set_caps (pad, caps);
}

/* chain function
 * this function does the actual processing
 */

static GstFlowReturn
gst_palmadpcm_chain (GstPad * pad, GstBuffer * buf)
{
  GstpalmADPCM *filter;
  GstBuffer *gstOutBuf;
  unsigned int outBuf[16334*2];
  unsigned long srcSize;
  unsigned long destSize;
  int err;

  GST_LOG("Chain\n");

  filter = GST_PALMADPCM (GST_OBJECT_PARENT (pad));

#if 0
  if(FALSE == filter->bRunning)
  {
        GST_LOG("OMG the filter is supposedly not running!\n");
  	gst_buffer_unref(buf);
	return GST_FLOW_ERROR;
  }
#endif

//  if (filter->silent == FALSE)
//    g_print ("I'm plugged, therefore I'm in.\n");

  if (filter->codec_initialized == FALSE) {
	GST_LOG("Initializing codec\n");
	// Ok, let's initialize the underlying codec session.
	filter->ADPCMParams.sampleRate = filter->rate;
	if(filter->channels == 1)
		filter->ADPCMParams.sampleWidth = palmCodecMONO;
	else
		filter->ADPCMParams.sampleWidth = palmCodecSTEREO;
	filter->ADPCMParams.blockSize = filter->block_align;
	filter->PCMParams.sampleRate = filter->rate;
	filter->PCMParams.sampleType = sndInt16Little;
	filter->PCMParams.sampleWidth = filter->ADPCMParams.sampleWidth;
	if(filter->adpcm_type == sndFormatDVI_ADPCM) {
		// Decode DVI ADPCM
		err = CodecCreateSession(palmCodecAudioIMA_ADPCM, 
			&filter->ADPCMParams,palmCodecAudioPCM, 
			&filter->PCMParams, 0xdeadbeef, &filter->pSession);
	}
	if(filter->adpcm_type == sndFormatIMA_ADPCM) {
		// Decode IMA ADPCM
		err = CodecCreateSession(palmCodecAudioIMA_ADPCM, 
			&filter->ADPCMParams, palmCodecAudioPCM, 
			&filter->PCMParams, 0xdeadbeef, &filter->pSession);
	}
	filter->codec_initialized = TRUE;
  }
  
//  Ok, grab the input data and feed it to the decoder.

  srcSize = GST_BUFFER_SIZE(buf);
  destSize = 16334*2;
  memset(&outBuf, 0, destSize);

  err = CodecEncodeDecode(filter->pSession, 
	GST_BUFFER_DATA(buf), &srcSize, &outBuf, &destSize);

  GST_LOG("CodecEncodeDecode %d\n", err);

  // outBuf now contains the decoded content.  We have to shove it into
  // a gStreamer buffer and pump it out...

  err = gst_pad_alloc_buffer_and_set_caps(filter->srcpad, 0, 
		destSize, GST_PAD_CAPS(filter->srcpad), &gstOutBuf);

  GST_LOG("alloc_buffer_and_set_caps %d\n", err);
  if(err != 0) {
	GST_LOG("Can't alloc, returning error\n");
	return err;
  }

  memcpy(GST_BUFFER_DATA(gstOutBuf), outBuf, destSize);
  GST_BUFFER_SIZE(gstOutBuf) = destSize;
  err = gst_pad_push(filter->srcpad, gstOutBuf);

  GST_LOG("gst_pad_push %d\n", err);

  gst_buffer_unref(buf);

  return GST_FLOW_OK;
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features
 *
 * exchange the string 'plugin' with your elemnt name
 */
static gboolean
plugin_init (GstPlugin * plugin)
{

  GST_LOG("***ADPCM plugin init\n");
  /* exchange the strings 'plugin' and 'Template plugin' with your
   * plugin name and description */
  GST_DEBUG_CATEGORY_INIT (gst_palmadpcm_debug, "palmadpcmdecoder",
      0, "ADPCM Decoder");

  gst_element_register (plugin, "palmadpcmdecoder",
      GST_RANK_PRIMARY, GST_TYPE_PALMADPCM);

#if 0
  // Removed for the moment
  if(!gst_type_find_register (plugin, "audio/x-adpcm", GST_RANK_MARGINAL,
	gst_palmadpcm_type_find, NULL, GST_CAPS_ANY, NULL, NULL)) {
	  printf("adpcm: failure to register type.");
  }
#endif
  
  return TRUE;  
}

#if 0
// Removed for the moment
static void gst_palmadpcm_type_find (GstTypeFind* tf, gpointer data)
{
	guint8* qcpString = gst_type_find_peek (tf, 8, 4);
	GstCaps* qcpCaps = NULL;

	GST_DEBUG ("In gst_palmadpcm_type_find with qcpString %s", qcpString);
//        if (strncmp (qcpString, "QLCM", 4) != 0)
//		return; /* probably not a qcelp file */

//	GST_DEBUG ("QLCM string matched ");
	qcpCaps = gst_caps_new_simple("audio/x-adpcm",
		"rate", G_TYPE_INT, 8000,
		"channels", G_TYPE_INT, 2,
		"width", G_TYPE_INT, 16, 
		"endianness", G_TYPE_INT, BYTE_ORDER,
		"depth", G_TYPE_INT, 16,
		"signed", G_TYPE_BOOLEAN, TRUE, NULL);
	
	gst_type_find_suggest (tf, GST_TYPE_FIND_LIKELY, qcpCaps);
	gst_caps_unref (qcpCaps);
}
#endif

/* this is the structure that gstreamer looks for to register plugins
 */
GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "adpcmdecoder",
    "Decode an IMA/ADPCM file",
    plugin_init, VERSION, "unknown", "Palm ADPCM Decoder", "Palm Media Division")
